/*
 TsumeCompe.js（Tsumeshogi Competition Tool）
    Ver 1.1 (2024-12-31)
     by KANEKO Kiyoshi / tsume-kobo.org & shogi-problem.org 2021-2024, CC-BY

    Ver 1.1  (2024-12-31)
    Ver 1.02（2023-05-05）
    Ver 1.01（2022-05-05）
    Ver 1.00（2021-11-23）
    Ver 0.70β(2021-05-07）

    Google Closure Compiler:  https://developers.google.com/closure/compiler
        Compile Option
        --compilation_level ADVANCED_OPTIMIZATIONS --language_in ECMASCRIPT_NEXT --language_out ECMASCRIPT_2021
*/
"use strict";

(() => { // GCCがvarを生成するため全体をクロージャにしておく

    // 共通の定数・関数
    const pxUnit = "px";
    const strictNumber = (p) => ((/\D/.test(p)) ? 0 : +p); // 数字だけの文字列から十進整数変換
    const noError = 0;
    const errorFileFormat = 1;
    const errorCsaHasMoves = 2;
    const errorOte = 3;

    const num7to1 = [7, 6, 5, 4, 3, 2, 1]; // 持駒文字列生成に使用するので7→1の順序
    const num9to1 = [9, 8, ...num7to1];

    const anyButtonY = (-3); // clickMap上の「１手もどる」等ボタンのY座標
    const defenseHandsY = (-1); // clickMap上の後手持駒表示のY座標
    const offenseHandsY = 11; // clickMap上の先手持駒表示のY座標
    const isDefenseHandsY = (y) => (y < 1); // 盤面上端より上
    const isOffenseHandsY = (y) => (y > 9); // 盤面下端より下

    // scriptタグに設定するパラメータ名称
    const parDateStart = "start"; // 開始時刻の設定
    const parDateEnd = "end"; // 終了時刻の設定
    const parIdPrefix = "idprefix"; // form-inputのID名称（answerに代える文字列）
    const parCookiePath = "cookie"; // cookieのパス。""ならcookie使用しない
    const parStartFile = "file"; // 問題ファイル番号の開始値及び拡張子指定。開始値は回答ID番号の開始値にも連動
    const parConfigFile = "config"; // 設定ファイルのファイル名。ファイルパラメータより優先
    const parRequiredId = "required"; // 必須項目のID名称（カンマ区切り列挙）とガイド表示名

    // 将棋盤関係。盤面表示と解答手順生成の兼用のため、ここに設置
    const pieceCharJ = [
        // 0     1     2     3     4     5     6     7
          "", "歩", "香", "桂", "銀", "金", "角", "飛", // 持駒など
        "玉",   "",   "",   "",   "",   "",   "",   "",
          "", "と", "杏", "圭", "全",   "", "馬", "龍",
          "",   "",   "",   "",   "",   "",   "",   "",
          "", "歩", "香", "桂", "銀", "金", "角", "飛", // 後手
        "王",   "",   "",   "",   "",   "",   "",   "",
          "", "と", "杏", "圭", "全",   "", "馬", "龍",
          "",   "",   "",   "",   "",   "",   "",   "",
          "", "歩", "香", "桂", "銀", "金", "角", "飛", // 先手
        "玉",   "",   "",   "",   "",   "",   "",   "",
          "", "と", "杏", "圭", "全",   "", "馬", "龍",
          "",   "",   "",   "",   "",   "",   "",   ""
    ];

    // 駒の動き。王手チェックと選択可能位置探索の兼用のため、ここに設置
    //　　0桂移動 1前 2前連続 3後横 4後横連続 5斜前 6斜後 7斜連続
    const pieceMoveKIN =
        [false,  true, false,  true, false,  true, false, false];
    const pieceMove = [
        [], // 0x00
        [false,  true, false, false, false, false, false, false], // 0x01 歩
        [false,  true,  true, false, false, false, false, false], // 0x02 香
        [ true, false, false, false, false, false, false, false], // 0x03 桂
        [false,  true, false, false, false,  true,  true, false], // 0x04 銀
        pieceMoveKIN, // 0x05 金
        [false, false, false, false, false,  true,  true,  true], // 0x06 角
        [false,  true,  true,  true,  true, false, false, false], // 0x07 飛
        [false,  true, false,  true, false,  true,  true, false], // 0x08 玉
        [], // 0x09
        [], // 0x0A
        [], // 0x0B
        [], // 0x0C
        [], // 0x0D
        [], // 0x0E
        [], // 0x0F
        [], // 0x10
        pieceMoveKIN, // 0x11 と
        pieceMoveKIN, // 0x12 成香
        pieceMoveKIN, // 0x13 成桂
        pieceMoveKIN, // 0x14 成銀
        [], // 0x15
        [false,  true, false,  true, false,  true,  true,  true], // 0x16 馬
        [false,  true,  true,  true,  true,  true,  true, false]  // 0x17 龍
    ];

    // 駒定数
    const pieceCodeFU = 1; // pieceCharToCode("FU");
    const pieceCodeKY = 2; // pieceCharToCode("KY");
    const pieceCodeKE = 3; // pieceCharToCode("KE");
    const pieceCodeGI = 4; // pieceCharToCode("GI"); // 現状未使用
    const pieceCodeKI = 5; // pieceCharToCode("KI");
    const pieceCodeKA = 6; // pieceCharToCode("KA"); // 現状未使用
    const pieceCodeHI = 7; // pieceCharToCode("HI"); // 現状未使用
    const pieceCodeOU = 8; // pieceCharToCode("OU");

    // 駒コード関係定数
    const bitMaskNari = 0x10;
    const bitMaskOffense = 0x20;
    const bitMaskDefense = 0x40;
    const pieceStone = (bitMaskOffense | bitMaskDefense); // 駒移動の壁（先手かつ後手）

    // 駒コード関係関数
    const isOffense = (c) => (c & bitMaskOffense);
    const isDefense = (c) => (c & bitMaskDefense);
    const isNotFromHand = (c) => (c & 0xf8);
    const toPieceOriginal = (c) => (c & 0x1f); // 成・不成を含めた駒種類

    // shogiBanまとめクロージャ
    const shogiBan = (csaText, isTestOte) => { //引数はcsaテキスト、王手チェックするか

        let currentBoardObj = {
            board: new Int8Array(81),
            offenseHand: new Int8Array(8),
            defenseHand: new Int8Array(8),
            offenseKingX: 0,
            offenseKingY: 0,
            defenseKingX: 0,
            defenseKingY: 0,
            nowMoves: 0,
            selectedToX: 0,
            selectedToY: 0,
            selectedFromX: 0,
            selectedFromY: 0,
            isNareru: false,
            ansString: ""
            // , prevObj:
        };

        let _maxMoves;
        let _errorId = noError;

        const pieceCharCsa = [
            //  0      1      2      3      4      5      6      7
            " * ",  "FU",  "KY",  "KE",  "GI",  "KI",  "KA",  "HI",
             "OU",    "",    "",    "",    "",    "",    "",    "",
               "",  "TO",  "NY",  "NK",  "NG",    "",  "UM",  "RY",
               "",    "",    "",    "",    "",    "",    "",    "",
               "", "+FU", "+KY", "+KE", "+GI", "+KI", "+KA", "+HI",
            "+OU",    "",    "",    "",    "",    "",    "",    "",
               "", "+TO", "+NY", "+NK", "+NG",    "", "+UM", "+RY",
               "",    "",    "",    "",    "",    "",    "",    "",
               "", "-FU", "-KY", "-KE", "-GI", "-KI", "-KA", "-HI",
            "-OU",    "",    "",    "",    "",    "",    "",    "",
               "", "-TO", "-NY", "-NK", "-NG",    "", "-UM", "-RY",
               "",    "",    "",    "",    "",    "",    "",    ""
        ];
        const pieceCharToCode = (c) => {
            const code = pieceCharCsa.indexOf(c);
            return (code < 0 ? 0 : code);
        }

        const _nowMoves = () => currentBoardObj.nowMoves;

        const _shogiBanIndex = (x, y) => ((x * 9) - y);
        const _toPieceOnHand = (c) => (c & 7); // 取ったときの駒種類

        const _incMoves = () => {
            currentBoardObj = {
                board: new Int8Array(currentBoardObj.board),
                offenseHand: new Int8Array(currentBoardObj.offenseHand),
                defenseHand: new Int8Array(currentBoardObj.defenseHand),
                offenseKingX: currentBoardObj.offenseKingX,
                offenseKingY: currentBoardObj.offenseKingY,
                defenseKingX: currentBoardObj.defenseKingX,
                defenseKingY: currentBoardObj.defenseKingY,
                nowMoves: currentBoardObj.nowMoves + 1,
                selectedToX: 0,
                selectedToY: 0,
                selectedFromX: 0,
                selectedFromY: 0,
                isNareru: false,
                ansString: "",
                prevObj: currentBoardObj
            };
        };

        const _toX = () => currentBoardObj.selectedToX;
        const _toY = () => currentBoardObj.selectedToY;

        const _decMoves = () => {
            if (_nowMoves()) { // nowMoves > 0
                currentBoardObj = currentBoardObj.prevObj;
                _errorId = noError;
            }
        };

        const _setAnsString = () => {
            const toX = _toX();
            const toY = _toY();
            const tmpFromY = currentBoardObj.selectedFromY;
            const fromY = (isDefenseHandsY(tmpFromY) || isOffenseHandsY(tmpFromY)) ? 0 : tmpFromY; // 打のとき0
            const fromX = fromY ? currentBoardObj.selectedFromX : 0; // 打のとき0
            const toXY = _shogiBanIndex(toX, toY);
            const ans = toX + (toY + pieceCharJ[currentBoardObj.board[toXY]] + fromX) + fromY;
            currentBoardObj.ansString = currentBoardObj.prevObj.ansString + " " + ans;
        };

        const _setFromCsa = () => {

            // Ver1.1　csa入力ファイルのチェックを全体的に強化
            const regBanmen = /^P[1-9]([\+\-](FU|KY|KE|GI|KI|KA|HI|OU|TO|NY|NK|NG|UM|RY)|[^\+\-]..){9}$/;; // 盤面
            const regDefenseHand =/^P\-(00(FU|KY|KE|GI|KI|KA|HI))+$/; // 後手持ち駒
            const regOffenseHand =/^P\+(00(FU|KY|KE|GI|KI|KA|HI))+$/; // 先手持ち駒
            // const regDefenseMove = /^\-(00|[1-9][1-9])[1-9][1-9](FU|KY|KE|GI|KI|KA|HI|OU|TO|NY|NK|NG|UM|RY)/; // 後手着手
            // const regOffenseMove = /^\+(00|[1-9][1-9])[1-9][1-9](FU|KY|KE|GI|KI|KA|HI|OU|TO|NY|NK|NG|UM|RY)/; // 先手着手

            for (const inTextLineTrim of csaText.split(/ *(\r\n|\n|,)/)) { // Ver1.1
                const inTextLinePad = inTextLineTrim.padEnd(29); // Ver1.1
                if (regBanmen.test(inTextLinePad)) {
                    const y = +inTextLinePad[1]; // .slice(1, 2);
                    for (const x of num9to1) { // Ver.1.02
                        const xc = 29 - (x * 3);
                        const k = pieceCharToCode(inTextLinePad.slice(xc, xc + 3));
                        currentBoardObj.board[_shogiBanIndex(x, y)] = k;
                        if (k == (bitMaskOffense | pieceCodeOU)) { // 先手玉
                            currentBoardObj.offenseKingX = x;
                            currentBoardObj.offenseKingY = y;
                        } else if (k == (bitMaskDefense | pieceCodeOU)) { // 後手玉
                            currentBoardObj.defenseKingX = x;
                            currentBoardObj.defenseKingY = y;
                        }
                    }
                } else if (regOffenseHand.test(inTextLineTrim)) {
                    const offenseHand = currentBoardObj.offenseHand;
                    for (let xc = 4; xc < inTextLineTrim.length; xc += 4) {
                        offenseHand[pieceCharToCode(inTextLineTrim.slice(xc, xc + 2))]++;
                    }
                } else if (regDefenseHand.test(inTextLineTrim)) {
                    const defenseHand = currentBoardObj.defenseHand;
                    for (let xc = 4; xc < inTextLineTrim.length; xc += 4) {
                        defenseHand[pieceCharToCode(inTextLineTrim.slice(xc, xc + 2))]++;
                    }
                } else if (/^[\+\-]../.test(inTextLineTrim)) { // Ver1.1。何らかの手順要素を含んでいるファイルとみなして警告表示
                    _errorId = errorCsaHasMoves;
                    return; // 強制終了
                } else if (/^'\* *\d+ *$/.test(inTextLineTrim)) {
                    _maxMoves ||= strictNumber(inTextLineTrim.slice(2).trim()); // コメントで表示手数を記入したとき
                }
            }
        };

        const _simpleMove = (toX, toY, fromX, fromY) => { // 持駒のときはfromXは駒種類
            currentBoardObj.selectedToX = toX;
            currentBoardObj.selectedToY = toY;
            currentBoardObj.selectedFromX = fromX;
            currentBoardObj.selectedFromY = fromY;
            const toXY = _shogiBanIndex(toX, toY);
            if (isDefenseHandsY(fromY)) { // defenseHandsY 後手の持駒
                currentBoardObj.board[toXY] = (fromX | bitMaskDefense);
                currentBoardObj.defenseHand[fromX]--;
            } else if (isOffenseHandsY(fromY)) { // offenseHandsY 先手の持駒
                currentBoardObj.board[toXY] = (fromX | bitMaskOffense);
                currentBoardObj.offenseHand[fromX]--;
            } else {
                const fromXY = _shogiBanIndex(fromX, fromY);
                const toK = currentBoardObj.board[toXY];
                const fromK = currentBoardObj.board[fromXY];
                const kcn = toPieceOriginal(fromK); // 成・不成コミの駒種類
                if (isOffense(fromK)) { // 先手番のとき
                    currentBoardObj.offenseHand[_toPieceOnHand(toK)]++;
                    currentBoardObj.board[fromXY] = 0;
                    if ((toY == 1) && ((kcn == pieceCodeFU) || (kcn == pieceCodeKY))) { // １段目の香歩
                        currentBoardObj.board[toXY] = (fromK | bitMaskNari);
                        currentBoardObj.isNareru = false;
                    } else if ((toY < 3) && (kcn == pieceCodeKE)) { // １・２段目の桂
                        currentBoardObj.board[toXY] = (fromK | bitMaskNari);
                        currentBoardObj.isNareru = false;
                    } else if (kcn == pieceCodeKI) { // 金
                        currentBoardObj.board[toXY] = fromK;
                        currentBoardObj.isNareru = false;
                    } else if (kcn == pieceCodeOU) { // 玉
                        currentBoardObj.board[toXY] = fromK;
                        currentBoardObj.isNareru = false;
                        currentBoardObj.offenseKingX = toX;
                        currentBoardObj.offenseKingY = toY;
                    } else if (((fromY < 4) || (toY < 4)) && !(fromK & bitMaskNari)) { // 成・不成選択。３段目以下、!は==0
                        currentBoardObj.board[toXY] = (fromK | bitMaskNari);
                        currentBoardObj.isNareru = true;
                    } else {
                        currentBoardObj.board[toXY] = fromK;
                        currentBoardObj.isNareru = false;
                    }
                } else { // 後手番のとき
                    currentBoardObj.defenseHand[_toPieceOnHand(toK)]++;
                    currentBoardObj.board[fromXY] = 0;
                    if ((toY == 9) && ((kcn == pieceCodeFU) || (kcn == pieceCodeKY))) { // ９段目の香歩
                        currentBoardObj.board[toXY] = (fromK | bitMaskNari);
                        currentBoardObj.isNareru = false;
                    } else if ((toY > 7) && (kcn == pieceCodeKE)) { // ８・９段目の桂
                        currentBoardObj.board[toXY] = (fromK | bitMaskNari);
                        currentBoardObj.isNareru = false;
                    } else if (kcn == pieceCodeKI) { // 金
                        currentBoardObj.board[toXY] = fromK;
                        currentBoardObj.isNareru = false;
                    } else if (kcn == pieceCodeOU) { // 玉
                        currentBoardObj.board[toXY] = fromK;
                        currentBoardObj.isNareru = false;
                        currentBoardObj.defenseKingX = toX;
                        currentBoardObj.defenseKingY = toY;
                    } else if (((fromY > 6) || (toY > 6)) && !(fromK & bitMaskNari)) { // 成・不成選択。７段目以上、!は==0
                        currentBoardObj.board[toXY] = (fromK | bitMaskNari);
                        currentBoardObj.isNareru = true;
                    } else {
                        currentBoardObj.board[toXY] = fromK;
                        currentBoardObj.isNareru = false;
                    }
                }
            }
            _setAnsString();
        };

        const _moveByClickXY = (toX, toY, fromX, fromY) => {
            _incMoves();
            _simpleMove(toX, toY, fromX, fromY);
            if (!isTestOte) return;
            // 以降は王手のチェック
            const isGote = isDefense(_pieceInBoard(toX, toY));
            const isOte = _isOte(true); // 後手玉王手チェック
            if (isGote) { // 後手番のとき（王手可否判定のみで、成・不成チャレンジ不要）
                _errorId = (isOte ? errorOte : noError);
            } else if (_isOte(false)) { // 先手玉王手チェック
                _errorId = errorOte;
            } else if (currentBoardObj.isNareru) { // 先手番で成・不成変更できるとき
                _changeNari(); // 成が合法かチャレンジ
                if (_isOte(true)) { // 成・不成変更後も後手玉が王手のとき
                    if (isOte)
                        _changeNari(); // 成・不成ともに合法・成に戻す
                    else {
                        currentBoardObj.isNareru = false; // 不成のみ合法
                    }
                } else {
                    if (isOte) {
                        _changeNari(); // 成のみ合法・成に戻す
                        currentBoardObj.isNareru = false;
                    } else {
                        _errorId = errorOte; // 成・不成ともに王手にならない
                    }
                }
            } else { // 先手番で成・不成変更できないとき
                _errorId = (isOte ? noError : errorOte);
            }
        };

        const _isOte = (isGoteGyoku) => {
            const testLong = (x, y, mx, my, kingSideBitMask, checkLong) => {
                const k = _pieceInBoard(x + mx, y + my);
                if (!k) { // ==0
                    return testLong(x + mx, y + my, mx, my, kingSideBitMask, checkLong);
                } else if (k & kingSideBitMask) { // !=0
                    return false;
                } else {
                    return pieceMove[toPieceOriginal(k)][checkLong];
                }
            };

            const testShort = (x, y, mx, my, kingSideBitMask, checkShort, checkLong) => {
                const k = _pieceInBoard(x + mx, y + my);
                if (!k) { // ==0
                    return testLong(x + mx, y + my, mx, my, kingSideBitMask, checkLong);
                } else if (k & kingSideBitMask) { // !=0
                    return false;
                } else {
                    return pieceMove[toPieceOriginal(k)][checkShort];
                }
            };

            const x = (isGoteGyoku ? currentBoardObj.defenseKingX : currentBoardObj.offenseKingX);
            const y = (isGoteGyoku ? currentBoardObj.defenseKingY : currentBoardObj.offenseKingY);
            if ((!x) || (!y)) { // x,yどちらか==0
                return false;
            }
            const forceY = (isGoteGyoku ? 1 : -1);
            const offenseBitMask = (isGoteGyoku ? bitMaskOffense : bitMaskDefense);
            const kingSideBitMask = (isGoteGyoku ? bitMaskDefense : bitMaskOffense);
            const offenseKE = (offenseBitMask | pieceCodeKE);

            return ((_pieceInBoard(x + 1, y + forceY + forceY) == offenseKE) ||
                (_pieceInBoard(x - 1, y + forceY + forceY) == offenseKE) ||
                (testShort(x, y, 0, forceY, kingSideBitMask, 1, 2)) || // 前方向
                (testShort(x, y, 0, -forceY, kingSideBitMask, 3, 4)) || // 後方向
                (testShort(x, y, 1, 0, kingSideBitMask, 3, 4)) || // 横方向
                (testShort(x, y, -1, 0, kingSideBitMask, 3, 4)) || // 横方向
                (testShort(x, y, 1, forceY, kingSideBitMask, 5, 7)) || // 斜め前方向
                (testShort(x, y, -1, forceY, kingSideBitMask, 5, 7)) || // 斜め前方向
                (testShort(x, y, 1, -forceY, kingSideBitMask, 6, 7)) || // 斜め後方向
                (testShort(x, y, -1, -forceY, kingSideBitMask, 6, 7))); // 斜め後方向
        };

        const _changeNari = () => { // 成・不成の反転、return 手順文字列
            currentBoardObj.board[_shogiBanIndex(_toX(), _toY())] ^= bitMaskNari; // 成ビットのXOR
            _setAnsString();
        };

        const _pieceInBoard = (x, y) => {
            if ((x < 1) || (x > 9) || (y < 1) || (y > 9)) {
                return pieceStone; // 駒移動の壁（先手かつ後手）
            } else {
                return currentBoardObj.board[_shogiBanIndex(x, y)];
            }
        };

        _setFromCsa();

        return {
            decMoves: _decMoves,
            nowMoves: _nowMoves,
            moveByClickXY: _moveByClickXY,
            changeNari: _changeNari,
            pieceInBoard: _pieceInBoard,
            offenseHand: (c) => currentBoardObj.offenseHand[_toPieceOnHand(c)],
            defenseHand: (c) => currentBoardObj.defenseHand[_toPieceOnHand(c)],
            selectedToX: _toX,
            selectedToY: _toY,
            isNareru: () => currentBoardObj.isNareru,
            errorId: () => _errorId,
            ansString: () => currentBoardObj.ansString,
            maxMoves: _maxMoves // 実質const
        };
    };

    // ダークモード関係まとめ
    const colorControls = (() => {
        let stickyTop = null;
        const elementsDrawList = [];
        const colorObj = {
            setElement: (elem) => elementsDrawList.push(elem.uiOption.fullDraw),
            setStickyTop: (elem) => {
                stickyTop = elem;
                elem.style.zIndex = 2;
                elem.style.opacity = "1.0";
                elem.style.position = "sticky";
                elem.style.top = "0" + pxUnit;
                onChangeDark();
            }
        };
        const darkThemeQuery = window.matchMedia("(prefers-color-scheme: dark)");
        const onChangeDark = () => {
            const isDark = darkThemeQuery.matches;
            if (stickyTop) {
                stickyTop.style.color = (isDark ? "white" : "black");
                stickyTop.style.backgroundColor = (isDark ? "black" : "white");
            }
            colorObj.fontColor = (isDark ? "white" : "black"); // 盤駒文字の色　デフォルト黒　ダークモードで白
            colorObj.highLightColor = (isDark ? "maroon" : "yellow"); // 着手地点の色　デフォルト黄色 ダークモードでマルーン
            colorObj.grayColor = (isDark ? "darkgray" : "lightgray"); // 選択していない盤面、選択失敗時の色
            for (const item of elementsDrawList) {
                item();
            }
        };
        darkThemeQuery.addEventListener("change", onChangeDark);
        onChangeDark();
        return colorObj;
    })();

    // uiOnShogiBanまとめクロージャ
    const uiOnShogiBan = (() => {
        // 駒以外の盤面表示文字
        const numberA = ["", "１", "２", "３", "４", "５", "６", "７", "８", "９"];
        const numberJ = ["", "一", "二", "三", "四", "五", "六", "七", "八", "九"];
        const noHandMark = "なし"; // なし
        const offenseHandMark = "☗"; // 先手駒形
        const defenseHandMark = "☖"; // 後手駒形

        // 駒を打てる位置のY座標
        const num9to3 = [9, 8, 7, 6, 5, 4, 3];
        const num9to2 = [...num9to3, 2];
        const num8to1 = [8, ...num7to1];
        const droppableRangeY = [
                            //   歩       香       桂       銀       金       角       飛
                            [[], num9to2, num9to2, num9to3, num9to1, num9to1, num9to1, num9to1], // 0先手
                            [[], num8to1, num8to1, num7to1, num9to1, num9to1, num9to1, num9to1]  // 1後手
                        ];

        const positionSet = (() => {
            // デフォルトフォント指定。monospace=スマホ版chromiumで不可、sans-serif=firefoxで不可
            const isFirefox = window.navigator.userAgent.toLowerCase().includes("firefox");
            const fontDefault = pxUnit + " " + (isFirefox ? "monospace" : "sans-serif");

            const szWidth = ((document.currentScript.parentNode.clientWidth || window.innerWidth) - 8) / 20; // 8=左右マージン
                                                                                                            // parentNodeのclientWidthがundefindedならばwindow.innerWidthから算出
            const szHeight = window.innerHeight / 36; // Ver1.1
            const sz = Math.floor(Math.min(szWidth, szHeight, 22));
            const leftMargin = Math.floor(sz / 3 + 1);
            const topMargin = Math.floor(sz / 2);

            const szScaleX = (p) => (sz * p) + leftMargin;
            const szScaleRoundX = (p) => Math.round(szScaleX(p));
            const szScaleY = (p) => (sz * p) + topMargin;
            const szScaleRoundY = (p) => Math.round(szScaleY(p));

            const tmpPathBoardGrid = new Path2D();
            const tmpFontPosX = new Int16Array(10);
            const tmpFontPosY = new Int16Array(10);
            const tmpGridPosX = new Int16Array(10);
            const tmpGridPosY = new Int16Array(10);

            const tmpFullGrid = sz * 2;
            const tmpButtonTopY = szScaleY(0);
            const tmpButtonBackLeftX = szScaleX(0);
            const tmpButtonNariLeftX = szScaleX(14);
            const tmpButtonWidth = sz * 4;

            const num9to0 = [...num9to1, 0];
            for (const xy of num9to0) {
                tmpFontPosX[xy] = szScaleX(19 - (2 * xy));
                tmpGridPosX[xy] = szScaleX(18 - (2 * xy));
                tmpFontPosY[xy] = szScaleY((2 * xy) + 7);
                tmpGridPosY[xy] = szScaleY((2 * xy) + 8);
            }

            for (const xy of num9to0) {
                tmpPathBoardGrid.moveTo(tmpGridPosX[xy], tmpGridPosY[0]);
                tmpPathBoardGrid.lineTo(tmpGridPosX[xy], tmpGridPosY[9]);
                tmpPathBoardGrid.moveTo(tmpGridPosX[0], tmpGridPosY[xy]);
                tmpPathBoardGrid.lineTo(tmpGridPosX[9], tmpGridPosY[xy]);
            }
            tmpPathBoardGrid.rect(tmpGridPosX[9], szScaleY(3), sz * 16, sz * 3);
            tmpPathBoardGrid.rect(tmpGridPosX[9], szScaleY(27), sz * 16, sz * 3);
            tmpPathBoardGrid.rect(tmpButtonBackLeftX, tmpButtonTopY, tmpButtonWidth, tmpFullGrid);
            tmpPathBoardGrid.rect(tmpButtonNariLeftX, tmpButtonTopY, tmpButtonWidth, tmpFullGrid);

            return {
                topMargin: topMargin,
                leftMargin: leftMargin,
                fullGrid: tmpFullGrid,
                fullWidth: sz * 20,
                fullHeight: sz * 31,

                offenseHandPosY: szScaleY(29),
                offenseHandCountPosY: szScaleRoundY(27.7), // 先手持駒数のY座標。標準27.5より少し下
                offenseHandHighLightY: szScaleY(28),
                defenseHandPosY: szScaleY(5),
                defenseHandCountPosY: szScaleRoundY(3.7), // 後手持駒数のY座標。標準3.5より少し下
                defenseHandHighLightY: szScaleY(4),

                noHandPosX: szScaleX(4),
                movesPosX: szScaleRoundX(17.5),
                movesWidth: sz * 3,

                fontPosX: tmpFontPosX,
                fontPosY: tmpFontPosY,
                gridPosX: tmpGridPosX,
                gridPosY: tmpGridPosY,

                buttonTopY: tmpButtonTopY,
                buttonBackLeftX: tmpButtonBackLeftX,
                buttonNariLeftX: tmpButtonNariLeftX,
                buttonWidth: tmpButtonWidth,

                buttonFontY: szScaleY(1),
                buttonBackFontX: szScaleX(2),
                buttonNariFontX: szScaleX(16),
                buttonFontWidth: Math.round(sz * 3.5),

                titleFontX: szScaleX(9),
                titleY: szScaleY(1),
                titleWidth: sz * 9,

                fontPosNumberX: szScaleRoundX(18.7), // 盤面漢数字のX座標。標準19より少し左
                fontPosNumberY: szScaleRoundY(7.3), // 盤面洋数字のY座標。標準7より少し下

                fontFull: Math.floor(sz * 1.4) + fontDefault, // スマホ用Firefox対応で少し小さめ
                fontHalf: Math.floor(sz * 0.8) + fontDefault,
                fontTitle: (sz * 2 + 1) + fontDefault,

                pathBoardGrid: tmpPathBoardGrid
            };

        })();

        // clickMap関係の定数
        const selectableFrom = 0x80;
        const removeSelectableFrom = (c) => (c & 0x7f); // 0x7fは!selectableFromの意味
        const selectableTo = 0x60; // == pieceStone
        const notSelectable = 0x68;
        const selectableBack = 0x70;
        const selectableNari = 0x78;
        const clickMapSizeX = 11;
        const clickMapSizeY = 17;
        const clickMapOffsetY = 4; // 将棋盤1段目が（各種ボタン含めた）clickMapの5段目
        const clickMapSize = clickMapSizeX * clickMapSizeY;
        const clickMapIndex = (x, y) => ((y * clickMapSizeX) + (clickMapOffsetY * clickMapSizeX) + x); // number型確定変数を先に計算した方が圧縮しやすい
        const initClickMap = new Int8Array(clickMapSize).fill(notSelectable); // clickMapの初期化値
        for (const x of num9to1) {
            for (const y of num9to1) {
                initClickMap[clickMapIndex(x, y)] = 0;
            }
        }

        return (myElement, canvasNumber) => {

            let titleString;
            let myShogiBan = { nowMoves: () => 0 }; // インスタンス化前にdrawBoardでnowMovesを判定させるダミー
            let selectedFromX = 0;
            let selectedFromY = 0;
            let selectedFromPiece = 0;
            let isSelectedFrom = false;
            let selectedMissX = 0;
            let selectedMissY = 0;
            let isSelectedMiss = false;
            let isInPointing = false;
            let downPointX = 0;
            let downPointY = 0;
            let firstDownPointX = 0;
            let firstDownPointY = 0;
            let clickMap; // 問題読み込み後に初期化する
            let isReadCsaComplete = false;

            const _isGoteban = () => (myShogiBan.nowMoves() & 1); // nowMovesが奇数のとき ==1

            const _getClickMap = (x, y) => clickMap[clickMapIndex(x, y)];
            const _setClickMap = (x, y, c) => (clickMap[clickMapIndex(x, y)] = c);

            const _readCsaText = (csaText, isTestOte) => {

                myShogiBan = shogiBan(csaText, isTestOte);
                if (myShogiBan.errorId()) { // ファイルの中に手順を含んでいる場合のみ
                    alert("第" + canvasNumber + "問は解答手順あり！");
                } else {
                    isReadCsaComplete = true;
                    titleString = "第" + canvasNumber + "問" +
                                    (myShogiBan.maxMoves
                                    ? "（" + myShogiBan.maxMoves + "手詰）" // maxMoves>0なら手数表示
                                    : "");
                    _fullDraw(false); // isNotFocus
                    colorControls.setElement(myElement);
                    setEvent(myElement);
                }
                return noError; // errorCsaHasMovesはリトライする意味がないのでnoErrorを返す
            };

            // canvas要素のスタイル初期化
            myElement.width = positionSet.fullWidth;
            myElement.height = positionSet.fullHeight;
            myElement.style.width = positionSet.fullWidth + pxUnit;
            myElement.style.height = positionSet.fullHeight + pxUnit;
            myElement.style.zIndex ||= 1; // z-indexは設定があれば優先（selectの規定値が２）
            const context = myElement.getContext("2d"); // canvasサイズ初期化後に取得

            const _fullDraw = (isFocus) => { // 全て描画
                _drawBoard(isFocus);
                if (isReadCsaComplete) {
                    _drawPiece();
                }
            };
            const fullDrawFalseFocus = () => _fullDraw(false);

            const _drawBoard = (isFocus) => { // 盤面の描画
                context.textAlign = "center"; // Ver1.1 再描画時にリセットされることがあるため
                context.textBaseline = "middle"; // Ver1.1 再描画時にリセットされることがあるため
                context.direction = "ltr"; // Candume Ver2.5相当の修正 rtl環境での利用に対応
                context.clearRect(0, 0, positionSet.fullWidth, positionSet.fullHeight);
                if (!isFocus) { // フォーカスでないとき
                    context.fillStyle = colorControls.grayColor;
                    context.fillRect(0, 0, positionSet.fullWidth, positionSet.fullHeight);
                }
                context.strokeStyle = colorControls.fontColor;
                if (isInPointing) {
                    const clickMapValue = _getClickMap(downPointX, downPointY);
                    if (clickMapValue != notSelectable) {
                        context.fillStyle = colorControls.grayColor;
                        if (clickMapValue == selectableBack) {
                            context.fillRect(positionSet.buttonBackLeftX, positionSet.buttonTopY, positionSet.buttonWidth, positionSet.fullGrid);
                        } else if (clickMapValue == selectableNari) {
                            context.fillRect(positionSet.buttonNariLeftX, positionSet.buttonTopY, positionSet.buttonWidth, positionSet.fullGrid);
                        } else if (isDefenseHandsY(downPointY)) { // defenseHandsY 後手の持駒
                            context.fillRect(positionSet.gridPosX[downPointX], positionSet.defenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                        } else if (isOffenseHandsY(downPointY)) { // offenseHandsY 先手の持駒
                            context.fillRect(positionSet.gridPosX[downPointX], positionSet.offenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                        } else {
                            context.fillRect(positionSet.gridPosX[downPointX], positionSet.gridPosY[downPointY - 1], positionSet.fullGrid, positionSet.fullGrid);
                        }
                    }
                } else if (isSelectedMiss) {
                    context.fillStyle = colorControls.grayColor;
                    if (isDefenseHandsY(selectedMissY)) { // defenseHandsY 後手の持駒
                        context.fillRect(positionSet.gridPosX[selectedMissX], positionSet.defenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                    } else if (isOffenseHandsY(selectedMissY)) { // offenseHandsY 先手の持駒
                        context.fillRect(positionSet.gridPosX[selectedMissX], positionSet.offenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                    } else {
                        context.fillRect(positionSet.gridPosX[selectedMissX], positionSet.gridPosY[selectedMissY - 1], positionSet.fullGrid, positionSet.fullGrid);
                    }
                }
                context.fillStyle = colorControls.highLightColor;
                if (isSelectedFrom) {
                    const psgx = positionSet.gridPosX[selectedFromX];
                    if (isDefenseHandsY(selectedFromY)) {
                        context.fillRect(psgx, positionSet.defenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                        context.strokeRect(psgx, positionSet.defenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                    } else if (isOffenseHandsY(selectedFromY)) {
                        context.fillRect(psgx, positionSet.offenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                        context.strokeRect(psgx, positionSet.offenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                    } else {
                        context.fillRect(psgx, positionSet.gridPosY[selectedFromY - 1], positionSet.fullGrid, positionSet.fullGrid);
                        context.strokeRect(psgx + 2, positionSet.gridPosY[selectedFromY - 1] + 2, positionSet.fullGrid - 4, positionSet.fullGrid - 4);
                    }
                } else if (myShogiBan.nowMoves()) { // >0
                    const xx = myShogiBan.selectedToX();
                    const yy = myShogiBan.selectedToY();
                    if (isDefenseHandsY(yy)) { // defenseHandsY 後手の持駒
                        context.fillRect(positionSet.gridPosX[xx], positionSet.defenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                    } else if (isOffenseHandsY(yy)) { // offenseHandsY 先手の持駒
                        context.fillRect(positionSet.gridPosX[xx], positionSet.offenseHandHighLightY, positionSet.fullGrid, positionSet.fullGrid);
                    } else {
                        context.fillRect(positionSet.gridPosX[xx], positionSet.gridPosY[yy - 1], positionSet.fullGrid, positionSet.fullGrid);
                    }
                }
                context.fillStyle = colorControls.fontColor;
                context.stroke(positionSet.pathBoardGrid);
                context.font = positionSet.fontHalf;
                for (const xy of num9to1) {
                    context.fillText(numberA[xy], positionSet.fontPosX[xy], positionSet.fontPosNumberY);
                    context.fillText(numberJ[xy], positionSet.fontPosNumberX, positionSet.fontPosY[xy]);
                }
                context.font = positionSet.fontFull;
                context.fillText(offenseHandMark, positionSet.fontPosX[9], positionSet.offenseHandPosY);
                context.fillText(defenseHandMark, positionSet.fontPosX[9], positionSet.defenseHandPosY);
            };

            const _drawPiece = (() => { // 駒の描画 + clickMap生成

                const _drawPieceOrg = () => {
                    const isGoteban = _isGoteban();
                    clickMap = new Int8Array(initClickMap); // clickMapをクリアする
                    context.fillStyle = colorControls.fontColor;
                    context.font = positionSet.fontFull;
                    for (const y of num9to1) {
                        for (const x of num9to1) {
                            const k = myShogiBan.pieceInBoard(x, y);
                            if (isOffense(k)) {
                                context.fillText(pieceCharJ[k], positionSet.fontPosX[x], positionSet.fontPosY[y]);
                                if (!isGoteban) { // 先手番のときだけ
                                    _setClickMap(x, y, (k | selectableFrom));
                                }
                            }
                        }
                    }
                    // 持駒の処理
                    let offenseHandColumn = 9;
                    let defenseHandColumn = 9;
                    for (const k of num7to1) {
                        const offenseHandCount = myShogiBan.offenseHand(k);
                        if (offenseHandCount) { // >0
                            const posX = positionSet.fontPosX[--offenseHandColumn]; // 表示カラムを１つ右
                            context.fillText(pieceCharJ[k], posX, positionSet.offenseHandPosY);
                            context.font = positionSet.fontHalf;
                            context.fillText(offenseHandCount > 9 ? offenseHandCount : numberA[offenseHandCount], // １ケタ数字は全角
                                posX, positionSet.offenseHandCountPosY);
                            context.font = positionSet.fontFull;
                            if (!isGoteban) { // 先手番のときだけ
                                _setClickMap(offenseHandColumn, offenseHandsY, (k | selectableFrom));
                            }
                        }
                        const defenseHandCount = myShogiBan.defenseHand(k);
                        if (defenseHandCount) { // >0
                            const posX = positionSet.fontPosX[--defenseHandColumn]; // 表示カラムを１つ右
                            context.fillText(pieceCharJ[k], posX, positionSet.defenseHandPosY);
                            context.font = positionSet.fontHalf;
                            context.fillText(defenseHandCount > 9 ? defenseHandCount : numberA[defenseHandCount], // １ケタ数字は全角
                                posX, positionSet.defenseHandCountPosY);
                            context.font = positionSet.fontFull;
                            if (isGoteban) { // 後手番のときだけ
                                _setClickMap(defenseHandColumn, defenseHandsY, (k | selectableFrom));
                            }
                        }
                    }
                    if (offenseHandColumn > 8) {
                        context.fillText(noHandMark, positionSet.noHandPosX, positionSet.offenseHandPosY);
                    }
                    if (defenseHandColumn > 8) {
                        context.fillText(noHandMark, positionSet.noHandPosX, positionSet.defenseHandPosY);
                    }

                    const mv = myShogiBan.nowMoves();
                    const mv2 = mv + (isSelectedFrom ? ("-" + (mv + 1)) : "手");
                    context.fillText(mv2, positionSet.movesPosX, positionSet.offenseHandPosY, positionSet.movesWidth);

                    context.scale(-1, -1); // 玉方駒は180度回転した盤面に負の座標で描画する
                    for (const y of num9to1) {
                        for (const x of num9to1) {
                            const k = myShogiBan.pieceInBoard(x, y);
                            if (isDefense(k)) {
                                context.fillText(pieceCharJ[k], -positionSet.fontPosX[x], -positionSet.fontPosY[y]); // 座標値は負の数
                                if (isGoteban) { // 後手番のときだけ
                                    _setClickMap(x, y, (k | selectableFrom));
                                }
                            }
                        }
                    }
                    context.scale(-1, -1); // 盤面回転を戻す
                    if (myShogiBan.nowMoves()) { // >0
                        context.fillText("１手もどる", positionSet.buttonBackFontX, positionSet.buttonFontY, positionSet.buttonFontWidth);
                        _setClickMap(9, anyButtonY, selectableBack);
                        _setClickMap(8, anyButtonY, selectableBack);
                    }
                    if (myShogiBan.isNareru()) {
                        context.fillText("成・不成", positionSet.buttonNariFontX, positionSet.buttonFontY, positionSet.buttonFontWidth);
                        _setClickMap(2, anyButtonY, selectableNari);
                        _setClickMap(1, anyButtonY, selectableNari);
                    }
                    context.font = positionSet.fontTitle;
                    context.fillText(titleString, positionSet.titleFontX, positionSet.titleY, positionSet.titleWidth);
                };

                const _setClickTo = () => {
                    const moveToXY = (x, y, bitMaskToForce) => { // １マス移動のとき
                        const k = myShogiBan.pieceInBoard(x, y);
                        if (!(k & bitMaskToForce)) { // ==0
                            _setClickMap(x, y, selectableTo);
                        }
                    };

                    const RunForXY = (orgX, orgY, mx, my, bitMaskToForce) => { // 連続移動のとき
                        const x = orgX + mx;
                        const y = orgY + my;
                        const k = myShogiBan.pieceInBoard(x, y);
                        if (!k) { // ==0
                            _setClickMap(x, y, selectableTo);
                            RunForXY(x, y, mx, my, bitMaskToForce);
                        } else if (!(k & bitMaskToForce)) { // ==0
                            _setClickMap(x, y, selectableTo);
                        }
                    };

                    const setDroppable = (x, y) => (myShogiBan.pieceInBoard(x, y) || _setClickMap(x, y, selectableTo));
                        // 駒打ち可能地点のセット
                        // if (myShogiBan.pieceInBoard(x, y))==0) _setClickMap(x, y, selectableTo);

                    const isGoteban = _isGoteban();
                    if (isNotFromHand(selectedFromPiece)) { // 持駒ではない
                        const forceY = (isGoteban ? 1 : -1); // 後手番ならY座標プラス向き、先手番ならY座標マイナス向きに探索
                        const wallToForce = (isGoteban ? (bitMaskDefense | pieceCodeOU) : (bitMaskOffense | pieceCodeOU)); // 進入不可能マス判定の壁（玉は４ビット目が１）

                        const selectedPieceMove = pieceMove[toPieceOriginal(selectedFromPiece)];
                        if (selectedPieceMove[0]) { // 桂
                            moveToXY(selectedFromX + 1, selectedFromY + forceY + forceY, wallToForce);
                            moveToXY(selectedFromX - 1, selectedFromY + forceY + forceY, wallToForce);
                            return;
                        }
                        if (selectedPieceMove[2]) { // 前連続
                            RunForXY(selectedFromX, selectedFromY, 0, forceY, wallToForce);
                        } else if (selectedPieceMove[1]) { // 前
                            moveToXY(selectedFromX, selectedFromY + forceY, wallToForce);
                        }
                        if (selectedPieceMove[4]) { // 横連続＋後連続
                            RunForXY(selectedFromX, selectedFromY, 1, 0, wallToForce);
                            RunForXY(selectedFromX, selectedFromY, -1, 0, wallToForce);
                            RunForXY(selectedFromX, selectedFromY, 0, -forceY, wallToForce);
                        } else if (selectedPieceMove[3]) { // 横＋後
                            moveToXY(selectedFromX + 1, selectedFromY, wallToForce);
                            moveToXY(selectedFromX - 1, selectedFromY, wallToForce);
                            moveToXY(selectedFromX, selectedFromY - forceY, wallToForce);
                        }
                        if (selectedPieceMove[7]) { // 斜連続。前後すべてテストするのでforceYでなく±1
                            RunForXY(selectedFromX, selectedFromY, 1, 1, wallToForce);
                            RunForXY(selectedFromX, selectedFromY, -1, 1, wallToForce);
                            RunForXY(selectedFromX, selectedFromY, 1, -1, wallToForce);
                            RunForXY(selectedFromX, selectedFromY, -1, -1, wallToForce);
                        } else if (selectedPieceMove[5]) { // 斜前
                            moveToXY(selectedFromX + 1, selectedFromY + forceY, wallToForce);
                            moveToXY(selectedFromX - 1, selectedFromY + forceY, wallToForce);
                            if (selectedPieceMove[6]) { // 斜後
                                moveToXY(selectedFromX + 1, selectedFromY - forceY, wallToForce);
                                moveToXY(selectedFromX - 1, selectedFromY - forceY, wallToForce);
                            }
                        }
                        return;
                    }
                    // 以下持駒の処理（持駒以外はreturn済）
                    const rangeY = droppableRangeY[isGoteban][selectedFromPiece];
                    if (selectedFromPiece == pieceCodeFU) { // 歩打ち
                        const targetNifu = (isGoteban ? (bitMaskDefense | pieceCodeFU) : (bitMaskOffense | pieceCodeFU));
                        for (const x of num9to1) {
                            if (rangeY.every((y) => myShogiBan.pieceInBoard(x, y) != targetNifu)) { // 二歩でないとき
                                for (const y of rangeY) {
                                    setDroppable(x, y);
                                }
                            }
                        }
                    } else { // 歩ではない持駒打ち
                        for (const y of rangeY) {
                            for (const x of num9to1) {
                                setDroppable(x, y);
                            }
                        }
                    }
                };

                return () => {
                    _drawPieceOrg();
                    if (isSelectedFrom) {
                        _setClickTo();
                    }
                }
            })();

            const _decMoves = () => {
                myShogiBan.decMoves();
                isSelectedFrom = false;
            };

            const _setXYbyClickPoint = (pt) => { // クリック位置xyによりdownPointX, downPointYを設定、値を変更したらtrueを返す
                const cx = pt.clientX;
                const cy = pt.clientY;
                const r = myElement.getBoundingClientRect();
                const x = Math.round(cx - r.left - positionSet.leftMargin); // leftMargin==クリックエリアの左端
                const y = Math.round(cy - r.top - positionSet.topMargin); // topMargin==クリックエリアの上端
                const fg = positionSet.fullGrid;
                const retx = 9 - Math.floor(x / fg);
                const rety = anyButtonY + Math.floor(y / fg);
                const modx = (x + 1) % fg;
                const mody = (y + 1) % fg;
                if (((modx < 3) || (mody < 3)) && (rety != anyButtonY)) {
                    downPointX = 0;
                    downPointY = 0;
                    return true;
                }
                if ((retx == downPointX) && (rety == downPointY)) {
                    return false;
                }
                downPointX = retx;
                downPointY = rety;
                return true;
            };

            const _setDownPoint = (pt) => {
                _setXYbyClickPoint(pt); // trueとfalse で同じ処理
                firstDownPointX = downPointX;
                firstDownPointY = downPointY;
                isInPointing = true;
                _fullDraw(true);
            };

            const _setMovePoint = (pt) => {
                if (!isInPointing) return;
                if (!_setXYbyClickPoint(pt)) return; // xy座標が変更ないとき
                const clickMapValue = _getClickMap(firstDownPointX, firstDownPointY);
                if (clickMapValue & selectableFrom) { // !=0。移動可能駒のドラッグ
                    selectedFromX = firstDownPointX;
                    selectedFromY = firstDownPointY;
                    selectedFromPiece = removeSelectableFrom(clickMapValue);
                    isSelectedFrom = true;
                    isSelectedMiss = false;
                }
                if (downPointY == anyButtonY) { // １手戻るとかのボタン位置（マス目相当のｙ座標）
                    downPointX = 0;
                    downPointY = 0;
                }
                firstDownPointX = 0;
                firstDownPointY = 0;
                _fullDraw(true);
            };

            const _setUpPoint = (pt) => {
                if (!isInPointing) return;
                isInPointing = false;
                if (_setXYbyClickPoint(pt)) return; // xy座標を変更したときreturn
                const clickMapValue = _getClickMap(downPointX, downPointY);
                switch (clickMapValue) {
                    case notSelectable:
                        isSelectedMiss = false; // セレクトミスをキャンセル
                        return; // Don't fullDraw
                    case selectableBack:
                        _decMoves();
                        isSelectedMiss = false; // セレクトミスをキャンセル
                        break;
                    case selectableNari:
                        myShogiBan.changeNari(); // 将棋盤クラスのchangeNariを呼び出す
                        isSelectedMiss = false; // セレクトミスをキャンセル
                        isSelectedFrom = false;
                        break;
                    case 0:
                        isSelectedMiss = true;
                        selectedMissX = downPointX;
                        selectedMissY = downPointY;
                        break;
                    case selectableTo:
                        if (selectedFromPiece < 8) {
                            myShogiBan.moveByClickXY(downPointX, downPointY, selectedFromPiece, selectedFromY);
                        } else {
                            myShogiBan.moveByClickXY(downPointX, downPointY, selectedFromX, selectedFromY);
                        }
                        if (myShogiBan.errorId()) { // 今のところ王手失敗のときだけ
                            _decMoves();
                            isSelectedMiss = true; // セレクトミスを設定
                            selectedMissX = downPointX;
                            selectedMissY = downPointY;
                            isSelectedFrom = true;
                        } else {
                            isSelectedFrom = false;
                            isSelectedMiss = false;
                        }
                        break;
                    default:
                        selectedFromX = downPointX;
                        selectedFromY = downPointY;
                        selectedFromPiece = removeSelectableFrom(clickMapValue);
                        isSelectedFrom = true;
                        isSelectedMiss = false;
                }
                _fullDraw(true);
            };

            _drawBoard(false); // isNotFocus

            document.addEventListener("visibilitychange", fullDrawFalseFocus); // candume Ver1.9 ブラウザの省メモリ機能で消去された盤面の再描画
            window.addEventListener("pageshow", fullDrawFalseFocus); // candume Ver1.9 ブラウザの省メモリ機能で消去された盤面の再描画

            return {
                elemNumber: canvasNumber,
                readCsaText: _readCsaText,
                fullDraw: _fullDraw,
                setDownPoint: _setDownPoint,
                setMovePoint: _setMovePoint,
                setUpPoint: _setUpPoint,
                setOutPointing: () => (isInPointing = false),
                ansString: () => myShogiBan.ansString()
            };
        };
    })();

    const setEvent = (() => { // イベント駆動のものをまとめて設定

        const curScriptDataset = document.currentScript.dataset;
        const docTarget = document.currentScript.parentNode; // DOMの処理範囲＝script設置場所の１段親
        docTarget.style.userSelect ||= "none"; // candumee Ver1.9相当 長押しで他要素が選択されるのを防止

        const targetCanvases = docTarget.getElementsByTagName("canvas");

        // イベントのキャンセル
        const tryCancel = (e) => {
            // マウス・タッチ・キーボードイベントのキャンセル処理
            // スクロール中はキャンセルはエラーになる等問題があるため、キャンセル処理を集約
            e.stopPropagation();
            if (e.cancelable) {
                e.preventDefault();
                return false;
            } else {
                return true;
            }
        };

        // 開始時刻と終了時刻の設定
        const getPlayDate = (dataName) => {
            const dateStringOrg = curScriptDataset[dataName];
            if (!dateStringOrg) return [new Date(0), false]; // 未指定と空文字列は同じ扱い
            // 文字列チェックを追加 Ver1.1
            const dateString = dateStringOrg.trim();
            const retDate = (() => {
                const regIsJST = /^\d{4}\-\d\d\-\d\dT\d\d\:\d\d\:\d\d$/;
                const regIsUTC = /^\d{4}\-\d\d\-\d\dT\d\d\:\d\d\:\d\d(\.\d{3})Z$/;
                const regIsCSV = /^\d{4}(\s*,\s*\d+)+$/;
                if (regIsJST.test(dateString)) {
                    return new Date(dateString + ".000+09:00"); // JST指定
                }
                if (regIsUTC.test(dateString)) {
                    return new Date(dateString); // UTC指定
                }
                const csvDate = (year, month, day = "01", hour = "00", minut = "00", second = "00", milli = "000") => {
                    const isoDate =     year.padStart(4, "0") + "-"
                                    +  month.padStart(2, "0") + "-"
                                    +    day.padStart(2, "0") + "T"
                                    +   hour.padStart(2, "0") + ":"
                                    +  minut.padStart(2, "0") + ":"
                                    + second.padStart(2, "0") + "."
                                    +  milli.padStart(3, "0") + "+09:00"; // JST
                    return new Date(isoDate);
                };
                if (regIsCSV.test(dateString)) { // カンマ区切り指定
                    return csvDate(...dateString.split(/\s*,\s*/)); // 空白除去
                }
                return NaN; // マッチしないのでエラー値を返す
            })();
            const errDate = isNaN(retDate); // 時刻生成エラーを判定
            if (errDate) alert("時刻指定誤り:" + dateString);
            return [retDate, !errDate];

        };
        const [dateStart, isDateStart] = getPlayDate(parDateStart);
        const [dateEnd, isDateEnd] = getPlayDate(parDateEnd);

        // ファイル番号と拡張子の設定
        const [startFileNumber, fileExtension] = (() => {
            const _startFile = ((curScriptDataset[parStartFile] || "1") + ".csa").split("."); // 未指定のときは"1.csa"とする
            const _StartNumber = strictNumber(_startFile[0]);
            const _startFileNumber = (_StartNumber || 1); // 指定が１以上の数値でないときは１
            const _fileExtension = "." + (_StartNumber ? _startFile[1] : _startFile[0]);
            return [_startFileNumber, _fileExtension];
        })();

        // formと送信ボタンの設定 Ver1.10でEnterキー誤送信防止を変更
        const usableForm = docTarget.getElementsByTagName("form")?.[0]; // formなしのHTMLも許容するためオプショナルチェーン
        const usableSubmitButton = (usableForm && docTarget.getElementsByTagName("button")?.[0]);  // 送信ボタンなしのHTMLも許容するためオプショナルチェーン
        if (usableSubmitButton) {
            usableSubmitButton.type = "button";
            usableSubmitButton.disabled = true;
            usableSubmitButton.addEventListener("keydown", tryCancel); // Ver1.10 ボタンFocus時のEnterキー抑止
            // 前行がダメなとき用に残しておく usableForm.addEventListener("keydown", (e) => (e.key == "Enter") && e.preventDefault() ); // formでのenterを無効化

            // Ver1.10 formの先頭にダミーボタンを設置しformのdefaultでのsubmitを防止
            const dummyButton =  document.createElement("button");
            dummyButton.type = "submit";
            dummyButton.style.display = "none";
            dummyButton.disabled = true;
            usableForm.insertBefore(dummyButton, usableForm.firstChild);
        }

        // 必須項目の設定。未指定のときは"name,place:氏名・居住地"とする
        const requiredString = (curScriptDataset[parRequiredId] || "name,place:氏名・居住地").split(":");
        const formRequired = [];
        for (const requierdId of requiredString[0].split(",")) { // 必須項目のelement
            const req = document.getElementById(requierdId.trim());
            if (!req) {
                alert("必須項目" + requierdId + "が見つかりません。");
                return;
            }
            req.value = "";
            formRequired.push(req);
        }
        let isRequiredComplete = false; // 必須項目の入力完了フラグ
        let reqGuide = requiredString[1]?.trim(); // ":"なし指定に備えオプショナルチェーン
        if (!reqGuide) { // 必須項目のガイドが指定されていないときlabel要素から取得
            const reqLabel = [];
            let isLabel = true;
            for (const elem of formRequired) {
                let labelText = "";
                for (const child of elem.parentNode.childNodes) {// 上位ノードをlabelとみなしてテキスト取得
                    if (child.nodeName == "#text"){
                        labelText = child.nodeValue.trim();
                        if (labelText) break;
                    }
                }
                if (labelText) {
                    reqLabel.push(labelText);
                } else {
                    isLabel = false;
                    break;
                }
            }
            if (isLabel) {
                reqGuide = reqLabel.join("・");
            }
        }
        const [msgBefore, msgAfter] = reqGuide && formRequired.length
            ? [reqGuide + "を入力して、開始時刻をお待ち下さい。", reqGuide + "を入力して下さい。"]
            : ["開始時刻までお待ち下さい。", ""];

        // メッセージ表示領域の関係
        const setGuideForce = (() => {
            const docsGuide = Array.from(docTarget.getElementsByTagName("div")).filter((elem) => !elem.innerHTML); // innerHTML==""
            if (!docsGuide.length) return (() => false); // 何もしない
            if ((docsGuide.length == 1) && (!docsGuide[0].style.position)) { // divが１つだけのときstickyにする
                const elem = docsGuide[0];
                colorControls.setStickyTop(elem);
                return (str) => elem.textContent = str;
            }
            return ((str) => {
                for (const elem of docsGuide) {
                    elem.textContent = str;
                }
            });
        })();

        const setGuideMsg = (() => {
            let msgPlayEnd = msgAfter;
            let latestMsg = "開始時刻までお待ちください。";

            return (msgString) => {
                latestMsg = msgString ?? latestMsg; // 空文字列は受け付ける
                const dateNow = new Date();
                if (isDateEnd && (dateNow > dateEnd)) {
                    if (isRequiredComplete) {
                        msgPlayEnd = "解答募集は終了しました。"
                    }
                    setGuideForce(msgPlayEnd);
                } else if (!isRequiredComplete) {
                    setGuideForce((isDateStart && (dateNow < dateStart)) ? msgBefore : msgAfter);
                } else {
                    setGuideForce(latestMsg);
                }
            };
        })();

        let isInPlay = null; // null → true → false に遷移
        const onInput = (() => {
            // リトライつきファイル読み込み～コールバックにテキストを渡す
            const callByReadText = (fileName, callback) => {
                const retryInterval = 5000; // 5秒 ファイル読み込みリトライ間隔
                const retryFetch = () => {
                    fetch(fileName, { cache: "no-store" }) // ファイル読み込みはキャッシュ不要のため
                        .then((response) => response.ok && response.text()) // response.ok==trueならばresponse.text()
                        .then((text) => {
                            if (!text || callback(text)) { // text=="" || callback!=noError
                                setTimeout(retryFetch, retryInterval); // 参加者側で対処不能なエラーは放置してリトライ
                            }
                        })
                        .catch((err) => setTimeout(retryFetch, retryInterval) ); // 参加者側で対処不能なエラーは放置してリトライ
                };
                retryFetch();
            };

            const startViewProblem = () => {
                let isConfigFile = false;
                let isTestOte;
                const readConfigText = (inText) => {
                    for (const text of inText.split(/\r\n|\n/)) {
                        switch (text.trim()) { // 拡張予定のためswitchにしておく
                            case "shokyu": {
                                isTestOte = true;
                                isConfigFile = true;
                                break;
                            }
                            case "ippan": {
                                isTestOte = false;
                                isConfigFile = true;
                                break;
                            }
                        }
                    }
                    if (!isConfigFile) {
                        return errorFileFormat;
                    }
                    for (const elem of targetCanvases) {
                        const fileName = elem.uiOption.elemNumber + fileExtension;
                        callByReadText(fileName, (csaText) => elem.uiOption.readCsaText(csaText, isTestOte));
                    }
                    return noError;
                };
                const configFileName = (curScriptDataset[parConfigFile] || ("0" + fileExtension)); // config未指定のときは"0.csa"とする
                callByReadText(configFileName, readConfigText);
            };

            let isWaitingViewProblem = true;

            const waitToViewProblem = () => {
                const displayTimerInterval = 10000; // 10秒
                const retryInterval = 5000; // 5秒
                const wait10Second = Math.floor((dateStart - (new Date())) / displayTimerInterval) * 10; // 10秒間隔の待ち時間表示
                if (isDateStart && (wait10Second > 0)) {
                    setGuideMsg("開始時刻まで約" + ((wait10Second < 600)
                                                    ? (wait10Second + "秒")
                                                    : (Math.floor(wait10Second / 60) + "分"))
                                                + "お待ち下さい。");
                    setTimeout(waitToViewProblem, retryInterval);
                } else {
                    setGuideMsg("混雑しています。このままお待ち下さい。");
                    startViewProblem();
                }
            };

            return () => {
                isRequiredComplete = formRequired.every((req) => req.value.trim()); // 必須項目すべてが空白でないとき req.value.trim()!="" // Ver1.02
                if (usableSubmitButton) {
                    usableSubmitButton.disabled = !isRequiredComplete && isInPlay;
                }
                setGuideMsg(); // 表示の更新のみ
                if (isRequiredComplete && isWaitingViewProblem) {
                    isWaitingViewProblem = false;
                    waitToViewProblem();
                }
            }
        })();

        const isTouchDevice = ("ontouchstart" in document); // タッチデバイスの判定（雑）

        // 送信処理関係
        const cancelUnload = (e) => { // ブラウザバックなどの誤操作にアラートを出す
            e.preventDefault();
            e.returnValue = "";
        };

        // フォームの回答input要素
        const inputPrefix = (curScriptDataset[parIdPrefix] || "answer"); // 未指定のときは"answer"とする
        const inputFormArea = (inputId) => document.getElementById(inputPrefix + inputId);

        // Cookie関係（使用可否判定、一意値生成、セット）
        const setCookie = () => {
            const _parCookie = curScriptDataset[parCookiePath];
            if (_parCookie == "") {  // cookiePathが""指定のときは、cookieを使わない（undefinedのときはcookieを使う）
                return true;
            }
            if (!navigator.cookieEnabled) {
                return false;
            }
            const uniqueValueDate =  Date.now() * (Math.random() + Math.random() + Math.random()) + "+" + Math.random(); // Ver1.10
            const cookieId = "TsumeCompe";
            const _cookiePath = ((_parCookie == "/") ? ";path=/" : ""); // "/"のときだけpath指定 v0.95で変更
            const _cookieSecure = ((location.protocol == "https:") ? ";secure" : "");
            const _cookieMaxAge = 40 * 24 * 60 * 60; // cookie寿命デフォルト40日
            const _cookieConst = ";samesite=strict;max-age=" + _cookieMaxAge + _cookieSecure + _cookiePath;
            const cookieInputArea = inputFormArea(0);

            const readCookie = () => {
                for (const cookieItem of document.cookie.split(";")) {
                    const p = cookieItem.split("=");
                    if (p[0].trim() == cookieId) {
                        return p[1].trim();
                    }
                }
                return uniqueValueDate; // readCookie()==""ならばuniqueValueを新調
            };

            const uniqueValue = readCookie();
            const cookieValue = cookieId + "=" + uniqueValue + _cookieConst;
            const writeCookie = () => (document.cookie = cookieValue);

            if (usableSubmitButton) {
                cookieInputArea.value = uniqueValue; // 重複チェックIDの格納
            }
            window.addEventListener("beforeunload", writeCookie);
            writeCookie();
            return (readCookie() == uniqueValue); // verify
        };

        let isSubmit = false;
        const onSubmit = () => {
            for (const elem of targetCanvases) { // フォームに解答書き込み
                elem.inputFormArea.value = elem.uiOption.ansString();
            }
            window.removeEventListener("beforeunload", cancelUnload);
            isSubmit = true;
            usableForm.submit(); // formなしのHTMLではonSubmitは発生しない
        };

        // UI element（canvas、input、selecct）にイベント設定
        const setUiEvent = (() => {

            // UI関係のイベント
            let currentElement = {}; // 操作中のcanvas要素
            let preSelectElement = {}; // 選択操作を開始したcanvas要素

            // マウス関係
            let isCancelMouseDown = false; // touchStart観測時に次のonDownを発火させない

            const onDown = (e) => {
                if (tryCancel(e) || isCancelMouseDown) {
                    isCancelMouseDown = false;
                    return;
                }
                if (!e.button) { // 左のみ処理 !は==0
                    if (e.target == currentElement) {
                        currentElement.uiOption.setDownPoint(e);
                    } else {
                        preSelectElement = e.target;
                    }
                }
            };

            const onMouseMove = (e) => {
                if (tryCancel(e)) return;
                if ((!e.button) && (e.target == currentElement)) { // 左のみ処理 !は==0
                    currentElement.uiOption.setMovePoint(e);
                }
            };

            const onUp = (e) => {
                if (tryCancel(e)) return;
                if (!e.button) { // 左のみ処理 !は==0
                    if (e.target == currentElement) {
                        currentElement.uiOption.setUpPoint(e);
                    } else if (e.target == preSelectElement) {
                        setFocus(preSelectElement);
                        preSelectElement = {};
                    }
                }
            };

            // タッチデバイス関係
            const onTouchStart = (e) => {
                isCancelMouseDown = true; // 直後のmouseDownをキャンセルする
                if (e.touches.length == 1) { // 指１本のみ処理
                    if (e.target == currentElement) {
                        currentElement.uiOption.setDownPoint(e.changedTouches[0]);
                    } else {
                        preSelectElement = e.target;
                    }
                }
            };

            let isCancelTouchEnd = false; // touchMove観測時に次のtouchEndを発火させない
            const onTouchEnd = (e) => {
                if (isCancelTouchEnd) { // touchMoveが発動した直後のtouchEnd → tryCancelしない
                    isCancelTouchEnd = false;
                    return;
                }
                tryCancel(e);
                if (!e.touches.length) { // ==0 指０本のみ処理
                    if (e.target == currentElement) {
                        currentElement.uiOption.setUpPoint(e.changedTouches[0]);
                    } else if (e.target == preSelectElement) {
                        setFocus(preSelectElement);
                        preSelectElement = {};
                    }
                }
            };

            // 問題選択関係
            const elementByNumber = []; // elementNumberからelementを取り出す　タッチデバイスのセレクタ用

            const usableSelectNo = ((selNo) => (isTouchDevice && (selNo.length == 1)) && selNo[0]
                )(Array.from(docTarget.getElementsByTagName("select")).filter( // タッチデバイス用の問題選択ボタンの探索
                    (elem) => {
                        let selectN = 0;
                        let nextN = startFileNumber;
                        for (const elemsOption of elem) {
                            if (elemsOption.value != selectN) {
                                return false;
                            }
                            selectN = (nextN++);
                        }
                        const selectNoStyle = elem.style;
                        selectNoStyle.visibility = "hidden"; // セレクトの初期状態は非表示
                        selectNoStyle.position = "fixed";
                        selectNoStyle.bottom ||= "0em"; // 設定があれば優先
                        selectNoStyle.right ||= "0em"; // 設定があれば優先
                        selectNoStyle.zIndex ||= 2; // 設定があれば優先（canvasの規定値が１）
                        return true;
                    }
                )
            );

            const onChangeSelect = () => { // ボタンオブジェクトを選択時にblurするのでcurrentのblurは不要
                const num = strictNumber(usableSelectNo.value);
                if (num) { // num!=0
                    setFocus(elementByNumber[num]);
                } else {
                    // num ==0 または NaN
                    window.scrollTo(0, 0);
                }
            };

            const setFocus = (elem) => ((elem != currentElement) || (elem != document.activeElement)) && elem.focus();

            const onFocus = (e) => {
                currentElement = e.target;
                if (usableSelectNo) {
                    usableSelectNo.value = currentElement.uiOption.elemNumber;
                }
                currentElement.uiOption.fullDraw(true);
                currentElement.scrollIntoView({
                    behavior: "smooth",
                    block: "center", // Ver1.1
                    inline: "nearest"
                });
            };

            const onBlur = (e) => {
                if (usableSelectNo) {
                    usableSelectNo.value = "-1"; // selectNo のエラー値を設定
                }
                currentElement = {};
                e.target.uiOption.fullDraw(false);
            };

            let unProcessFilesCount = targetCanvases.length; // 未処理数の初期値

            return (elem) => {

                if (isTouchDevice) {
                    elem.addEventListener("touchstart", onTouchStart, { passive: true });
                    elem.addEventListener("touchend", onTouchEnd);
                    elem.addEventListener("touchmove", (e) => (isCancelTouchEnd = true), { passive: true });
                    elementByNumber[elem.uiOption.elemNumber] = elem;
                }
                elem.addEventListener("mousedown", onDown);
                elem.addEventListener("mouseup", onUp);
                elem.addEventListener("mousemove", onMouseMove);
                elem.addEventListener("mouseleave", (e) => e.target.uiOption.setOutPointing() );
                elem.addEventListener("focus", onFocus);
                elem.addEventListener("contextmenu", tryCancel); // 右クリック抑止
                elem.addEventListener("selectstart", tryCancel); // 長タップ抑止（candume Ver1.2相当の修正）
                elem.addEventListener("click", tryCancel); // 削除できるかも知れない
                elem.addEventListener("dblclick", tryCancel); // 削除できるかも知れない
                elem.addEventListener("blur", onBlur);
                elem["tabIndex"] = 0;
                elem.inputFormArea = inputFormArea(elem.uiOption.elemNumber);

                if (--unProcessFilesCount) { // unProcessFilesCount!=0 →未処理のcanvasあり
                    return;
                }
                // window.addEventListener("beforeunload", cancelUnload); // Ver1.01で削除

                usableSubmitButton?.addEventListener("click", onSubmit); // 送信ボタンなしのHTMLも許容するためオプショナルチェーン
                isInPlay ??= true; // nullならばtrue
                if (usableSelectNo) {
                    usableSelectNo.addEventListener("change", onChangeSelect);
                    usableSelectNo.style.visibility = "visible";
                }
                if (isDateEnd) {
                    endTimer();
                } else {
                    setGuideMsg(""); // 表示の終了
                }
            }
        })();

        const endTimer = () => {
            const oneSecond = 1000;
            const oneMinute = oneSecond * 60;
            const oneHour = oneMinute * 60;
            const oneDay = oneHour * 24;
            const remain = dateEnd - new Date();
            if (remain < 0) {
                usableSubmitButton.disabled = true;
                isInPlay = false;
                setGuideMsg(); // 終了表示
            } else if (remain < oneMinute) { // 1分未満
                const s = Math.floor(remain / oneSecond);
                setGuideMsg("終了時刻まで" + s + "秒です。");
                setTimeout(endTimer, 100); // 0.1秒ごと更新
            } else if (remain < oneHour) { // 1時間未満
                const m = Math.floor(remain / oneMinute);
                const s = Math.floor((remain - m * oneMinute) / oneSecond);
                setGuideMsg("終了時刻まで" + m + "分" + s + "秒です。");
                setTimeout(endTimer, oneSecond); // 1秒ごと更新
            } else if (remain < oneDay) { // 1日未満
                const h = Math.floor(remain / oneHour);
                const m = Math.floor((remain - h * oneHour) / oneMinute);
                setGuideMsg("終了時刻まで約" + h + "時間" + m + "分です。");
                setTimeout(endTimer, oneMinute); // 1分ごと更新
            } else { // 1日以上
                const d = Math.floor(remain / oneDay);
                const h = Math.floor((remain - d * oneDay) / oneHour);
                const m = Math.floor((remain - d * oneDay - h * oneHour) / oneMinute);
                setGuideMsg("終了時刻まで約" + d + "日" + h + "時間" + m + "分です。");
                setTimeout(endTimer, oneMinute); // 1分ごと更新
            }

        };

        return (elem) => {
            if (elem) { // !=undefined
                setUiEvent(elem);
            } else if (setCookie()) {
                const forceReload = () => {
                    window.removeEventListener("beforeunload", cancelUnload);
                    location.reload();
                };
                window.addEventListener("pageshow", (e) => {
                    // Ver1.1 ページ不正遷移の対応を修正
                    // ブラウザによって挙動が異なるため、一律にリロードする
                    //if (e.persisted) { // キャッシュから表示されたとき→ダメ
                    //    forceReload();
                    //    return;
                    //}
                    //const entries = performance.getEntriesByType("navigation");
                    //for (const entry of entries) {
                    //    if (entry.type == "back_forward") { // ブラウザバックで表示されたとき→ダメ
                    //        forceReload();
                    //        return;
                    //    }
                    //};
                    if (isSubmit) { // submitが済んでる。他に方法がないときの最後の手段
                        forceReload();
                        return;
                    }

                });
                window.removeEventListener("beforeunload", cancelUnload);
                window.addEventListener("beforeunload", cancelUnload);
                let canvasCount = startFileNumber;
                for (const elem of targetCanvases) {
                    elem.uiOption = uiOnShogiBan(elem, canvasCount++); // Ver1.1
                }
                for (const req of formRequired) {
                    req.addEventListener("input", onInput); // 必須項目すべてにイベント設定
                }
                onInput(); // メッセージ欄の初期設定
            } else {
                setGuideForce("Cookieを有効にして再度接続して下さい。"); // cookie
            }
        };
    })();

    setEvent(); // パラメータをundefで呼び出し

})();

